from pathlib import Path

from .utils.singleton import Singleton
from .utils.yaml import CommentedMap

DEFAULTS = [
    ("device", str, "cpu"),
    ("output_dir", Path, "output"),
    ("experiment_name", str, None),
    ("log_dir", Path, "log"),
]


class Env(dict, metaclass=Singleton):
    def __init__(self):
        ...

    @staticmethod
    def from_config(cls, config: CommentedMap):
        assert isinstance(config, CommentedMap), type(config)
        items = []
        for key, tp, default in DEFAULTS:
            if key not in config:
                value = default
            else:
                value = tp(config.pop(key))
            items.append((key, value))
        items += list(config.items())



from ..utils import Singleton


class ClassRegistry(metaclass=Singleton):
    def __init__(self):
        self._classes = {}

    def register(self, cls, name=None):
        if name is None:
            name = cls.__name__

        if cls not in self._classes:
            self._classes[cls] = {}

        for c in cls.mro():
            if c in self._classes:
                assert name not in self._classes[c], f"Duplicate class name. {c} {name}"
                self._classes[c][name] = cls

    def get(self, cls, name=None):
        if name is None:
            name = cls.__name__

        if name not in self._classes[cls]:
            return None
        else:
            return self._classes[cls][name]

    def __contains__(self, cls):
        return cls in self._classes


def register(cls=None, name=None):
    def wrapper(x):
        ClassRegistry().register(x, name=name)
        return x

    if cls is None:
        return wrapper
    else:
        return wrapper(cls)



import functools
from abc import ABCMeta, abstractmethod
from typing import Any, List, Union

from ..utils.yaml import CommentedMap, CommentedSeq, TaggedScalar
from .registry import ClassRegistry
from .type_utils import get_full_arg_spec

GenericAlias = type(List)
NoneType = type(None)

__all__ = [
    "ConfigSpec",
    "ScalarConfigSpec",
    "OptionalConfigSpec",
    "TuplesConfigSpec",
    "ListConfigSpec",
    "ClassConfigSpec",
    "FunctionConfigSpec",
]


class ConfigSpec(metaclass=ABCMeta):
    @staticmethod
    def from_type(tp):
        if tp in (int, float, bool, str):
            return ScalarConfigSpec(tp)
        elif isinstance(tp, GenericAlias):
            origin = tp.__origin__
            if origin is list:
                internal_type = tp.__args__[0]
                return ListConfigSpec(internal_type)
            elif origin is Union:
                internal_types = tp.__args__
                if len(internal_types) == 2 and (
                    internal_types[0] is NoneType or internal_types[1] is NoneType
                ):
                    # Optional[T]
                    internal_type = (
                        internal_types[0]
                        if internal_types[0] is not NoneType
                        else internal_types[1]
                    )
                    return OptionalConfigSpec(internal_type)
            elif origin is tuple:
                internal_types = tp.__args__
                return TuplesConfigSpec(internal_types)
            else:
                raise NotImplementedError(tp)
        elif tp in ClassRegistry():
            return ClassConfigSpec(tp)
        else:
            raise NotImplementedError(tp)

    @abstractmethod
    def instantiate(self, config):
        ...


class ScalarConfigSpec(ConfigSpec):
    def __init__(self, tp: Any):
        self._tp = tp

    def instantiate(self, config: Any):
        return self._tp(config)


class OptionalConfigSpec(ConfigSpec):
    def __init__(self, tp: Any):
        self._internal_spec = ConfigSpec.from_type(tp)

    def instantiate(self, config: Any):
        if config is None:
            return None
        else:
            return self._internal_spec.instantiate(config)


class TuplesConfigSpec(ConfigSpec):
    def __init__(self, types: List[Any]):
        self._internal_specs = [ConfigSpec.from_type(tp) for tp in types]

    def instantiate(self, config: CommentedSeq):
        assert isinstance(config, CommentedSeq), config
        assert len(config) == len(self._internal_specs), (
            len(config),
            len(self._internal_specs),
        )
        return tuple(
            spec.instantiate(cfg) for spec, cfg in zip(self._internal_specs, config)
        )


class ListConfigSpec(ConfigSpec):
    def __init__(self, internal_type: Any):
        self.internal_spec = ConfigSpec.from_type(internal_type)

    def instantiate(self, config: CommentedSeq):
        assert isinstance(config, CommentedSeq), config
        return list(map(lambda x: self.internal_spec.instantiate(x), config))


class ClassConfigSpec(ConfigSpec):
    def __init__(self, cls):
        self._cls = cls

    @staticmethod
    @functools.lru_cache(maxsize=None)
    def get_argument_config_specs(cls):
        specs = {}

        arguments_spec, defaults = get_full_arg_spec(cls)
        for key, tp in arguments_spec.items():
            specs[key] = ConfigSpec.from_type(tp)

        return specs, defaults

    def instantiate(self, config: Union[str, TaggedScalar, CommentedMap]):
        params = {}
        if isinstance(config, str):
            # config is a class name
            cls = ClassRegistry().get(self._cls, config)
            assert cls is not None, f"{self._cls} {config}"
        elif isinstance(config, TaggedScalar):
            # config is a class name
            name = config.value[1:]  # strip off `!`
            cls = ClassRegistry().get(self._cls, name)
            assert cls is not None, f"{self._cls} {name}"
        elif isinstance(config, CommentedMap):
            tag = config.tag
            if tag.value is not None:
                name = tag.value[1:]  # strip off `!`
                cls = ClassRegistry().get(self._cls, name)
                assert cls is not None, f"{self._cls} {name}"
            else:
                cls = self._cls

            specs, defaults = self.get_argument_config_specs(cls)
            for key in config.keys():
                if key not in specs:
                    raise RuntimeError(f"unknown key: {key} ({cls})")
            for key, spec in specs.items():
                if key in config:
                    params[key] = spec.instantiate(config[key])
                elif key in defaults:
                    params[key] = defaults[key]
                else:
                    raise RuntimeError(f"Missing required key: {key} ({cls})")

        else:
            raise NotImplementedError(config, type(config))

        return cls(**params)


class FunctionConfigSpec(ConfigSpec):
    def __init__(self):
        pass



import inspect

__all__ = ["get_full_arg_spec"]


def get_full_arg_spec(cls):
    assert inspect.isclass(cls), cls

    argspec = inspect.getfullargspec(cls.__init__)
    # skip `self`
    args = argspec.args[1:] + argspec.kwonlyargs

    defaults = {}
    if argspec.defaults is not None:
        for k, v in zip(argspec.args[-len(argspec.defaults) :], argspec.defaults):
            defaults[k] = v
    if argspec.kwonlydefaults is not None:
        defaults.update(argspec.kwonlydefaults)

    annotations = argspec.annotations

    spec = {}
    for arg in args:
        assert arg in annotations, f"{cls} -> {arg}"
        spec[arg] = annotations[arg]

    return spec, defaults



from typing import Any, Dict


class Singleton(type):
    _instances: Dict[Any, Any] = {}

    def __call__(cls):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__()
        return cls._instances[cls]



__all__ = [
    "YAML",
    "CommentedMap",
    "CommentedSeq",
    "CommentedSet",
    "CommentedOrderedMap",
    "CommentedBase",
    "Tag",
    "TaggedScalar",
]

try:
    from ruamel.yaml import YAML
    from ruamel.yaml.comments import (
        CommentedBase,
        CommentedMap,
        CommentedOrderedMap,
        CommentedSeq,
        CommentedSet,
        Tag,
        TaggedScalar,
    )
except ModuleNotFoundError:
    from ruamel_yaml import YAML
    from ruamel_yaml.comments import (
        CommentedBase,
        CommentedMap,
        CommentedOrderedMap,
        CommentedSeq,
        CommentedSet,
        Tag,
        TaggedScalar,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sotalab', 'sotalab.core', 'sotalab.core.config', 'sotalab.core.utils']

package_data = \
{'': ['*']}

install_requires = \
['ruamel.yaml>=0.16.12,<0.17.0']

setup_kwargs = {
    'name': 'sotalab-core',
    'version': '0.1.3',
    'description': 'SotaLab',
    'long_description': '# SotaLab-Core\n',
    'author': 'Ming Yang',
    'author_email': 'ymviv@qq.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
